/**
 * Merge objects together, even their sub objects
 * @param {Object|Array<Object>} object - The object(s) to merge into the target
 * @param {Object} target - The target having all of its variables changed
 * @returns {Object}
 */

function deepMerge(object, target) {
    if (getType(target) !== 'object') throw new TypeError('The target should be an object to merge into');
    if (getType(object) !== 'object' && getType(object) !== 'array') throw new TypeError('The object should be an object, map or an array of objects/maps');
    if (object instanceof Array && target) return object.map(obj => deepMerge(obj, target));

    for (let index = 0; index < Object.keys(object).length; index++) {
        let keys = Object.keys(object),
            key = keys[index];

        if (getType(object[key]) === 'undefined' || getType(object[key]) === 'null') continue;

        switch (getType(object[key])) {
            case 'object':
                if (getType(target[key]) !== 'object') target[key] = target[key] ? Object.assign({}, target[key]) : {};
                deepMerge(object[key], target[key]);
                break;

            case 'array':
                if (getType(target[key]) !== 'array') target[key] = target[key] ? [target[key]] : [];
                for (let i = 0; i < object[key].length; i++) {
                    const value = object[key][i];
                    if (!target[key].includes(value)) target[key].push(value)
                }
                break;

            case 'map':
                if (getType(target[key]) === 'map') {
                    for (const [k, v] of object[key]) {
                        target[key].set(k, v)
                    }
                }
                break;

            default: target[key] = object[key];
        }
    }

    return target;
}

/**
 * Get the proper type of a JavaScript construct
 * @param {*} x - The js constructor to get the type of
 * @returns {string}
 */

function getType(x) {
    if (typeof x === 'string' || x instanceof String) return 'string';
    if (typeof x === 'number' && isFinite(x)) return 'number';
    if (x && typeof x === 'object' && x.constructor === Array) return 'array';
    if (typeof x === 'function') return 'function';
    if (x && typeof x === 'object' && x.constructor === Object) return 'object';
    if (x instanceof Map) return 'map';
    if (x === null) return 'null';
    if (typeof x === 'undefined') return 'undefined';
    if (typeof x === 'boolean') return 'boolean';
    if (x && typeof x === 'object' && x.constructor === RegExp) return 'regexp';
    if (x instanceof Date) return 'date';
    if (typeof x === 'symbol') return 'symbol';
    return typeof x;
}

module.exports = deepMerge;